<?php

require_once 'inc/regexfun.php';

/*

Handles expressions:

 [expr:EXPR]

Expressions may not be nested.

Expressions are actually PHP expressions with some restrictions:
No functions may be called or variables referenced.
Only a few constants are defined - they're interpreted
as calls to functions of the site.

*/

function WikiExpressionParse($expression, &$error)
{
  $matches = Array();
  preg_match_all
  (
    '@([[:alpha:]]\w*|-?[0-9]+|!=|>=|<=|==|[()!<>]|&&|\|\|)@',
    $expression,
    $matches
  );
  $eval_expr = '';
  $parens = 0;
  $par1_ok = true;
  $was_elem = false;
  foreach($matches[0] as $token)
  {
    switch($token)
    {
      /* BOOL functions - used by IF/ENDIF code */
      /* NUMERIC functions - used by EXPR code */
      case 'time':
        $token .= '()';
        break;
    }
    if($token != 'true'
    && $token != 'false'
    && preg_match('@^[[:alpha:]]\w*$@', $token))
    {
      $error = '<<undefined function "'.$token.'" in "'.$expression.'">>';
      return;
    }
    if($token == '(')
    {
      if(!$par1_ok) continue;
      ++$parens;
    }
    if($token == ')')
    {
      --$parens;
    }
    $is_elem = preg_match('@[[:alnum:]\w]@', $token[0]);
    if($is_elem && $was_elem)
    {
      $error = '<<"'.$expression.'": syntax error>>';
    }
    $par1_ok = !$is_elem;
    $was_elem = $is_elem;
    $eval_expr .= $token;
  }
  if($parens != 0)
  {
    $error = '<<"'.$expression.'": unbalanced parenthesis>>';
    return;
  }
  $cond = @eval('return '.$eval_expr.';');
  return $cond;
}

function WikiExpressionConvert($expr)
{
  if($GLOBALS['WIKI_PARSER'] != WIKI_PARSING_NORMAL) return '';
  
  $expr = WikiUnescapeLink($expr);
  
  $error = '';
  $result = WikiExpressionParse($expr, $error);
  if($error) return $error;
  return $result;
}

function WikiExpressionsProcess($text)
{
  return preg_replace
  (
    '@\[expr:((?:[^\]])*)\]@e',
    "WikiExpressionConvert(str_replace('\\\"', '\"', '\\1'))",
    $text
  );
}
