<?php

require_once 'inc/email.php';
require_once 'inc/dbfun.php';
require_once 'inc/wikidefs.php';
require_once 'inc/wikirender.php';
require_once 'inc/wikicond.php';
require_once 'inc/wikiexpr.php';
require_once 'inc/wikitabs.php';
require_once 'inc/smallcache.php';

function IsOkNewWikiFileName($name)
{
  if($name == 'endif') return false;
  return true;
}
function WikiTitleToWikiFile($pagetitle)
{
  $pagetitle = str_replace(Array('?','!','-','&','|','_','.','\''), Array('','',' ','','',' ',' ',''), $pagetitle);
  
  $base = preg_replace('@^ *(.*[^ ]) *$@', '\1',
    preg_replace('@ +(/?) ?([a-zA-Z])@e', "'\\1'.strtoupper('\\2')", ' '.$pagetitle));
  
  $base = preg_replace('/ /', '', $base);
  
  $num = 0;
  $result = $base;
  while(!IsOkNewWikiFileName($result) && $num<100)
  {
    $result = $base.++$num;
  }
  return $result;
}

function WikiFileToWikiTitle($filename)
{
  if(preg_match('/^[0-9]*S$/', $filename))
  {
    return 'Submission #'.(int)$filename;
  }
  if(preg_match('/^[0-9]*M$/', $filename))
  {
    return 'Movie #'.(int)$filename;
  }
  return preg_replace('@^ *([^ ].*[^ ]) *$@U', '\1',
    preg_replace('@([^ ]) */ *([^ ])@e', "str_replace('\\\"', '\"', '\\1 / \\2')",
    preg_replace('@([a-z])([A-Z0-9])@e', "str_replace('\\\"', '\"', '\\1 \\2')",
    preg_replace('@([A-Z0-9])([^A-Z0-9 ])@e', "str_replace('\\\"', '\"', ' \\1\\2')",
                $filename.' '))));
}

function WikiEscapeLink($link)
{
  return str_replace(Array('[',']'),Array('[[',']]'), $link);
}
function WikiUnescapeLink($link)
{
  return str_replace(Array('[[',']]'),Array('[',']'), $link);
}

function WikiLinkLinkHtmlFix($text)
{
  /* Create something that can be used in <a href="here">.
   * This means that " must be converted to &quot;
   * but &quot; must not be converted to &amp;quot;
   */
  return htmlspecialchars(WikiBriefUnHTML($text));
}
function WikiLinkTextHtmlFix($text)
{
  /* Convert wikiparser html.. to html. */
  return $text;
}

function WikiProduceRelativeLink($link)
{
  if($GLOBALS['WIKI_URLS'] == WIKI_URLS_SNAPSHOT)
    $link = SnapshotTranslateLink($link);

  if(!preg_match('@^(git:|http:|ftp:|ed2k:|irc:|nntp:|news:|/)@', $link))
  {
    if($GLOBALS['WIKI_URLS'] == WIKI_URLS_ABSOLUTE)
      $link = SITE_ABSOLUTE_BASE.$link;
    else
      $link = FixRelativeLink($link);
  }
  
  return $link;
}

function WikiLinkFunc($text, $cutter=true, $autoembedimg=false)
{
  #$text = HTMLtoWikiText($text);
  
  $link = $text;

  $was_cut = false;
  if($cutter && preg_match('@\|@', $text))
  {
    $link = preg_replace('@\|.*@', '', $text);
    $text = substr($text, strlen($link)+1);
    $was_cut = true;
  }

  if($GLOBALS['WIKI_PARSER'] == WIKI_PARSING_TEXT)
  {
    return $text;
  }

  if(preg_match('/mailto:/', $link))
  {
    if($text == $link) $text = ''; else $text .= ' ';
    $link = substr($link, 7);
    return GetEmail($text, '', $link);
  }
  
  $adefs = '';
  
  if(preg_match('@^=@', $link))
  {
    /* It's a direct reference */
    $link = substr($link, 1);

    if($GLOBALS['WIKI_PARSER'] == WIKI_PARSING_LINKS)
    {
      #print "link:$link\n";
      if(preg_match('@^queue\.cgi\?id=[0-9]+$@', $link))
      {
        $linktmp = ((int)str_replace('queue.cgi?id=', '', $link)) . 'S';
        ++$GLOBALS['WIKI_LINKS'][$linktmp];
      }
      elseif(preg_match('@^movies\.cgi\?id=[0-9]+$@', $link))
      {
        $linktmp = ((int)str_replace('movies.cgi?id=', '', $link)) . 'M';
        ++$GLOBALS['WIKI_LINKS'][$linktmp];
      }
    }

    if(!preg_match('@^(git:|http:|ftp:|ed2k:|irc:|nntp:|news:|/)@', $link))
    {
      $link = WikiProduceRelativeLink($link);
    }
    $text = WikiLinkTextHtmlFix($text);
  }
  elseif(!preg_match('@^(git:|http:|ftp:|ed2k:|irc:|nntp:|news:|/)@', $link))
  {
    /* It's a wiki page */
    
    # Convert it to a wikiname
    switch($GLOBALS['WIKI_PARSER'])
    {
      case WIKI_PARSING_LINKS:
      case WIKI_PARSING_TEXT:
        if(preg_match('@^(module:|if:|endif$|expr:)@', $link))
          return WikiLinkTextHtmlFix($link);
        break;
    }
    
    if(preg_match('@^module:@', $link))
    {
      if($text == $link) $text = '';
      return WikiBuiltInModule(substr($link, 7), $text);
    }
    if(preg_match('@^user:@', $link))
    {
      $link = substr($link, 5);
      if($was_cut) $link .= '|'.$text;
      $text = $link;
      $hp = false;//UserLinkHomepage($link);
      if(!$hp)
      {
        return WikiLinkTextHtmlFix($text);
      }
      $link = $hp;

      if($GLOBALS['WIKI_PARSER'] == WIKI_PARSING_LINKS)
      {
        ++$GLOBALS['WIKI_LINKS'][$link];
      }
    }
    else
    {
      $link = WikiTitleToWikiFile($link);
      
      $wikilink = preg_replace('/#.*/', '', $link);
      
      if($GLOBALS['WIKI_PARSER'] == WIKI_PARSING_LINKS)
      {
        ++$GLOBALS['WIKI_LINKS'][$wikilink];
      }
      
      $newlink = $wikilink;//ConvertHomepageLink($wikilink);
      if($newlink != $wikilink)
      {
        /* The page was changed - take the #xx part from
         * the old link to new link
         */
        $link = $newlink . substr($link, strlen($wikilink));
      }
    }

    $exists = WikiCheckPageExists($link);
    if(!$exists)
    {
      $adefs .= ' class="broken"';
    }

    $link .= '.html';
    
    $link = WikiProduceRelativeLink($link);
    $text = WikiLinkTextHtmlFix($text);
  }
  else
  {
    if(preg_match('@\.(png|jpg|gif)$@', $link) && !$autoembedimg)
    {
      $align = '';
      if($text == 'left' || $text == 'right') $align = $text;
      
      if($GLOBALS['WIKI_URLS'] == WIKI_URLS_SNAPSHOT)
      {
        $link = SnapshotTranslateImage($link);
      }
      
      $result = '<img src="'. WikiLinkLinkHtmlFix($link). '"';
      switch($align)
      {
        case 'left': $result .= ' class="embedleft"'; break;
        case 'right': $result .= ' class="embedright"'; break;
        default: $result .= ' class="embed"'; break;
      }
      $result .= ' alt="'.$text.'"';
      $result .= '/>';
      
      $text = $result;
    }
    else
    {
      $text = WikiLinkTextHtmlFix($text);
      if(/*$GLOBALS['WIKI_URLS'] != WIKI_URLS_ABSOLUTE
      &&*/ ($GLOBALS['WIKI_PARSER'] == WIKI_PARSING_NORMAL
         || $GLOBALS['WIKI_PARSER'] == WIKI_PARSING_SNAPSHOT)
      && 
      !preg_match('@^210\.150\.160\.@', $_SERVER['REMOTE_ADDR']) // *.excite.co.jp
      )
      {
        #$imgurl = FixRelativeLink('css/extlink.gif');
        #$text = '<img src="'.$imgurl.'" class="extlink" alt="(ext)"/> '.$text;
        $adefs .= ' class="extlink"';
        #$imgurl = FixRelativeLink('css/extlink.gif');
        #$text = '<img src="'.$imgurl.'" class="extlink" alt="(ext)"/> '.$text;
        
        if($GLOBALS['WIKI_PARSER'] != WIKI_PARSING_SNAPSHOT
        && !preg_match('@TASvideos@', $link))
        {
          $adefs .= ' id="'.sprintf('i%X', crc32(WikiBriefUnHTML($link).SQLPASS)).'"';
          $GLOBALS['HAS_EXTERNAL_LINKS'] = true;
        }
      }
    }
  }

  if(preg_match('@^(wiki|diff)\.cgi@', $link))
  {
    $adefs .= ' rel="nofollow"';
  }
  
  return '<a'.$adefs.' href="'.WikiLinkLinkHtmlFix($link).'">'.$text.'</a>';
}


function WikiGetPageTypeOf($pagetitle)
{
  // Pure numbers are MOVIE indices.
  if(preg_match('@^[0-9]+M?$@', $pagetitle)) return 'M';
  if(preg_match('@^[0-9]+S$@', $pagetitle)) return 'S';
  // Others are just PAGES
  return 'P';
}

function WikiIsSystemPage($pagedata)
{
  if(preg_match('@^System@', $pagedata['pagename'])
  && $pagedata['pagename'] != 'SystemPages')
  {
    $SQL = 'select * from site_links l,site_text t'.
           ' where t.pagename="SystemPages"'.
           ' and l.text_id=t.id'.
           ' and l.dest_page='.sqlfix($pagedata['pagename']).
           ' and t.obsoleted_by=-1';
    if(Fetch1Row($SQL)) return true;
  }
  return false;
}

function WikiIsDeletedPage($pagedata)
{
  if(preg_match('@^DeletedPages/@', $pagedata['pagename']))
  {
    return true;
  }
  return false;
}

function WikiCheckPageExists($pagename)
{
  static $cache;
  if(!$cache)
  {
    /* Cache the list of all existing pages */
    $SQL = 'select distinct upper(pagename)p from site_text'.
           ' where obsoleted_by=-1';
    $cache = Array();
    for($tmp = DbExecute($SQL); !$tmp->EOF; $tmp->MoveNext())
      $cache[$tmp->fields['p']] = true;
  }
  return array_key_exists(strtoupper($pagename), $cache);
}

function WikiCreateEditURL($pagename)
{
  #return '[' . $pagename . ']';
  return 'wiki.exe?page='.urlencode($pagename).'&mode=edit';
}

// Note: This function is called also by WikiProcessTOC()
function WikiBuiltInModuleDoExec($func, $params)
{
  $paramlist = explode('|', trim($params));
  $params = Array();
  foreach($paramlist as $param)
  {
    $key = preg_replace('@=.*@', '', $param);
    $value = substr($param, strlen($key)+1);
    $params[$key] = $value;
  }
  return $func($params);
}

function WikiBuiltInModule($modulename, $params = '')
{
  $modulename = preg_replace('/[^a-z0-9]/', '', strtolower($modulename));
  $modulename = preg_replace('/^[0-9]+/',   '', $modulename);
  
  $modulefn = "modules/{$modulename}.php";
  if(file_exists($modulefn))
  {
    require_once $modulefn;
    
    $cachevarname = "{$modulename}_cacheable";
    if(isset($GLOBALS[$cachevarname])
    && $GLOBALS['WIKI_PARSER'] != WIKI_PARSING_SNAPSHOT)
    {
      $cachedepvar = "{$modulename}_cachedepsql";
      $timestamp = filemtime($modulefn);
      $expiretime = $GLOBALS[$cachevarname];
      
      if($modulename == 'moviechangelog' && preg_match('/float=1/', $params))
      {
        $expiretime = 0; // no caching because of ads
      }
      
      if(isset($GLOBALS[$cachedepvar]))
      {
        $tmp = Fetch1Row($GLOBALS[$cachedepvar]);
        foreach($tmp as $k=>$last_update);
        if($last_update > $timestamp) $timestamp = $last_update;
      }
      
      if($modulename == 'wikitextchangelog' || $modulename == 'moviechangelog')
      {
        $params .=  '|timezone='.PrefGetTimezone();
      }
      
      if($expiretime > 0)
      {
        $version = $timestamp;
        return eaccelerator_cache_result
          (md5("nesvmodule:$version:$modulename:$params"),
           'WikiBuiltInModuleDoExec($modulename, $params)',
           $expiretime);
      }
    }
    return WikiBuiltInModuleDoExec($modulename, $params);
  }
  return htmlspecialchars('<<Module "'.$modulename.'" does not exist>>');
}

function GetLoadWikiPageSQL($pagename, $revision=0)
{
  $SQL = 'select * from site_text';
  $SQL .= ' where pagename='.sqlfix($pagename);
  if($revision)
  {
    $SQL .= ' and revision='.sqlfix($revision);
  }
  return $SQL;
}

function ResolveWikiPageAlias($pagename)
{
  $SQL = 'select actualname from page_alias where pagename='.sqlfix($pagename);
  $tmp = DbExecute($SQL);
  if(!$tmp->EOF)
    return $tmp->fields['actualname'];
}

function LoadWikiPage($pagename, $revision=0)
{
/* - Dangerous!
  $alias = ResolveWikiPageAlias($pagename);
  if($alias)
    $pagename = $alias;
*/
  $SQL = GetLoadWikiPageSQL($pagename, $revision);
  if($revision == 0)
    $SQL .= ' and obsoleted_by=-1';

  $SQL .= ' order by revision desc';
  $SQL .= ' limit 1';

  $tmp = DbExecute($SQL);
  if($tmp->EOF)
  {
    $desc = 'Page not found, and page [SystemPageNotFound] was not found either.';
    if($pagename != 'SystemPageNotFound')
    {
      $data = LoadWikiPage('SystemPageNotFound');
      $desc = $data['description'];
    }
    $data = Array
    (
      'description' => $desc,
      'pagename' => $pagename,
      'timestamp' => 0,
      'revision' => 1,
      'userid' => 0,
      'found' => false
    );
  }
  else
  {
    $data = $tmp->fields;
    $data['found'] = true;
    unset($tmp);
  }
  return $data;
}

function CalcPageSearchMetaphone($word)
{
  return preg_replace('/[^A-Z]/','',strtr(strtoupper($word), 'CERZDGUY', 'KIASTKAI'));
}
function GenPageSearchTitleFrom($pagename)
{
  switch(WikiGetPageTypeOf($pagename))
  {
    case 'S':
    {
      static $subnames;
      if(!$subnames) $subnames = GetSubmissionNameList();
      $id = (int)$pagename;
      if(!array_key_exists($id,$subnames)) return "? #$id";
      return $subnames[$id];
    }
    case 'M':
    {
      static $movnames;
      if(!$movnames) $movnames = GetMovieNameList();
      $id = (int)$pagename;
      if(!array_key_exists($id,$movnames)) return "? [$id]";
      return $movnames[$id];
    }
    case 'P':
    default:
      return WikiFileToWikiTitle($pagename);
  }
}
function GenMovieTitleMetaphone($movieid)
{
  static $movnames;
  if(!$movnames) $movnames = GetMovieNameList();
  $result = '';
  foreach(explode(' ', $movnames[$movieid]) as $word)
    $result .= ' '.CalcPageSearchMetaphone($word);
  return $result; 
}
function GenMovieClassMetaphone($movieid)
{
  $classlist = GetMovieClassList($movieid);
  $result = '';

  /* ignore some very common terms (damage Y/N, fastest, Npassword, Nsave) */
  $ignore = "/2Y/2N/3Y/7N/8N/";
  foreach($classlist as $class)
  {
    $cid = $class['id'];
    $pol = $class['value'];
    $c = $cid.($pol > 0 ? 'Y' : 'N');
    if(preg_match("@/$c/@", $ignore)) continue;
    $result .= " C$c";
  }
  return $result; 
}
function GenSearchTitleMetaphoneFrom($pagename)
{
  switch(WikiGetPageTypeOf($pagename))
  {
    case 'S':
    case 'P':
    case 'M':
    default:
    {
      $result = '';
      foreach(explode(' ', GenPageSearchTitleFrom($pagename)) as $word)
        $result .= ' '.CalcPageSearchMetaphone($word);
      return $result;
    }
  }
}
function GenSearchClassMetaphoneFrom($pagename)
{
  switch(WikiGetPageTypeOf($pagename))
  {
    case 'M':
    {
      return GenMovieClassMetaphone((int)$pagename);
    }
    case 'S':
    {
      $subid = (int)$pagename;
      $SQL = 'select id from movie where submissionid='.sqlfix($subid);
      $tmp = DbExecute($SQL);
      if(!$tmp->EOF) return GenMovieClassMetaphone($tmp->fields['id']);
      //passthru
    }
    case 'P':
    default: return '';
  }
}

function WikiPostParseText($textid)
{
  $SQL = 'delete from site_links where text_id='.sqlfix($textid);
  DbExecute($SQL);
  
  $SQL = 'select description,pagename,obsoleted_by from site_text where id='.sqlfix($textid);
  $tmp = DbExecute($SQL);
  if($tmp->EOF) return;
  
  if($tmp->fields['obsoleted_by'] > 0) return;
  
  $GLOBALS['WIKI_PARSER'] = WIKI_PARSING_LINKS;
  $GLOBALS['WIKI_LINKS'] = Array();
  
  $desc = $tmp->fields['description'];

  /* Update the search data */
  $search_title     = GenPageSearchTitleFrom($tmp->fields['pagename']);
  $title_metaphone = GenSearchTitleMetaphoneFrom($tmp->fields['pagename']);
  $class_metaphone = GenSearchClassMetaphoneFrom($tmp->fields['pagename']);
  $SQL = 'replace into site_text_search'.
         '(site_text_id,title,title_metaphone,class_metaphone,description)'.
         'values('.$textid.
               ','.sqlfix($search_title).
               ','.sqlfix($title_metaphone).
               ','.sqlfix($class_metaphone).
               ','.sqlfix($desc).')';
  DbExecute($SQL);

  WikiTextToHTML($desc); // find links in the page.
  
  $links = Array();
  foreach($GLOBALS['WIKI_LINKS'] as $link => $count)
  {
    if($link == '') continue;
    
    $type = WikiGetPageTypeOf($link);
    if($type == 'P')
    {
      // fix the capitalization of the link if possible
      $SQL = 'select pagename from site_text where pagename='.sqlfix($link);
      $tmp = Fetch1Row($SQL);
      if($tmp) $link = $tmp['pagename'];
    }
    
    $links[$link] += $count;
  }
  
  $multi = Array();
  foreach($links as $link => $count)
    $multi[] = Array('text_id' => $textid, 'link_count' => $count, 'dest_page' => $link);
  
  if(!empty($multi))
  {
    $SQL = GetMultiInsertSQL('site_links', $multi);
    DbExecute($SQL);
  }
  
  if($desc != '' && !preg_match("@\n$@", $desc)) $desc .= "\n";
  $nlines = substr_count($desc, "\n");
  
  $SQL = 'update site_text set num_lines='.sqlfix($nlines).
         ' where id='.sqlfix($textid);
  DbExecute($SQL);

  $GLOBALS['WIKI_PARSER'] = WIKI_PARSING_NORMAL;
}

function WikiUpdateSiteText($pagename, $new_text, $time,$userid,
                            $minoredit="N",
                            $whyedit='',
                            $old_rev=0)
{
  require_once 'inc/ipfun.php';
  
  $SQL = 'select revision,id,userid,timestamp,description'.
         ' from site_text where pagename='.sqlfix($pagename);
  $PageLoadSQL = $SQL;
  $SQL .= ' and obsoleted_by=-1 order by revision desc limit 1';
  
  $current_text = '';
  
  $tmp = Fetch1Row($SQL);
  $current_rev = 0;
  $current_id = 0;

  if($minoredit == 'N' && WikiGetPageTypeOf($pagename) != 'S')
  {
    $username = UserGetNameById($userid);
    $verb = $tmp ? 'edited' : 'created';
    NesVideoAgentAnnounceMessage(
      "Page $pagename $verb by $username ($whyedit): http://tasvideos.org/{$pagename}.html"
    );
  }

  if($tmp)
  {
    $current_rev  = $tmp['revision'];
    $current_id   = $tmp['id'];
    $current_user = $tmp['userid'];
    $current_time = $tmp['timestamp'];
    $current_text = $tmp['description'];
    
    if($current_user == $userid
    && $current_time >= $time-COMBINE_WIKIEDITS_PERIOD)
    {
      /* Combine the previous version to this one */
      $fields = Array
      (
        'timestamp'   => $time,
        'description' => $new_text,
        'minoredit'   => $minoredit,
        'whyedit'     => $whyedit,
        'ipaddr'      => GetIPaddr()
      );
      $SQL = GetUpdateSQL('site_text', $fields, $current_id);
      DbExecute($SQL);
      
      WikiPostParseText($current_id);
      return Array('id' => $current_id, 'errors' => 0, 'text' => $new_text);
    }
  }

  /* Inserting a new version */
  $type = WikiGetPageTypeOf($pagename);
  $fields = Array
  (
    'userid'      => $userid,
    'timestamp'   => $time,
    'description' => $new_text,
    'type'        => $type,
    'minoredit'   => $minoredit,
    'whyedit'     => $whyedit,
    'ipaddr'      => GetIPaddr(),
    'pagename'    => $pagename,
    'revision'    => $current_rev + 1
  );
  
  $nerrors = 0;
  
  /* FIXME: This doesn't work if the user brought in &rev= in the URL */
  if(false && $old_rev && $current_rev && $old_rev != $current_rev)
  {
    /* The version the user has edited is not the same as
     * we found in the database.
     * Must do a merging. */
    
    /* Take a patch from old and submitted version,
     * and apply that patch to the current.
     */
    
    /* FIXME TODO */
    $SQL = $PageLoadSQL . ' and revision='.sqlfix($old_rev);
    $tmp = Fetch1Row($SQL);
    $old_text = $tmp->fields['description'];
    
    /* Create a patch between old and new */
    $f1 = '/tmp/diff1'.getmypid(); file_put_contents($f1, $old_text);
    $f2 = '/tmp/diff2'.getmypid(); file_put_contents($f2, $new_text);
    $f3 = '/tmp/diff3'.getmypid(); file_put_contents($f3, $current_text);
    $f4 = '/tmp/diff4'.getmypid();
    $nerrors = (int)exec("diff3 -m $f2 $f1 $f3 > $f4; echo $?");
    $new_text = file_get_contents($f4);
    unlink($f1);
    unlink($f2);
    unlink($f3);
    unlink($f4);
    $fields['description'] = $new_text;
  }
  switch($type)
  {
    case 'P': break;
    case 'M': $fields['movieid']  = (int)$pagename; break;
    case 'S': $fields['submissionid'] = (int)$pagename; break;
  }
  $fields['id'] = GenIDfor('site_text');
  $SQL = GetInsertSQL('site_text', $fields);
  DbExecute($SQL);
  
  #print_r($fields);
  #exit;
  
  WikiPostParseText($fields['id']);
  
  if($current_id)
  {
    $SQL = 'update site_text set obsoleted_by='.$fields['id'].' where id='.$current_id;
    DbExecute($SQL);
  }
  
  return Array('id' => $fields['id'],
               'errors' => $nerrors,
               'text' => $new_text
              );
}


function UserGetWikiName()
{
  return WikiTitleToWikiFile(UserGetName());
}

function LoadWikiPageHTML($pagename)
{
  $tmp = LoadWikiPage($pagename);
  return WikiTextToHTML($tmp['description']);
}

function WikiGetPageLanguageCode($pagename)
{
  if(preg_match('@\\.pt@', $pagename)) return 'pt-BR';
  if(preg_match('@\\.ja@', $pagename)) return 'ja-JP';
  return 'en-US';
} 
      
function WikiLoadJavascript($filename, $inplace = false)
{
  if($inplace)
  {
    return '<script type="'.htmlspecialchars('text/javascript').'"'.
                   ' src="'.htmlspecialchars($filename).'"></script>';
  }
  AddPageIncludes($filename, 'text/javascript');
  return '';
}
function WikiEmbedJavascript($javascript, $defer=false)
{
  # AddPageEmbeds($javascript, 'text/javascript', $defer);
  # return '';
  
  $defer = $defer ? 'defer="defer"' : '';
  return "<script type='text/javascript' $defer><!--\n$javascript\n--></script>";
  #return "<script type='text/javascript' $defer><![CDATA[\n$javascript\n]]></script>";
}
