<?php

require_once 'inc/wikidefs.php';
require_once 'inc/smallcache.php';
require_once 'inc/wikisrcembed.php';
require_once 'inc/wikitoc.php';

$GLOBALS['WIKI_PARSER'] = WIKI_PARSING_NORMAL;
$GLOBALS['WIKI_URLS']   = WIKI_URLS_LOCAL;
$GLOBALS['strdyn'] = Array();
$GLOBALS['strcon'] = Array();
$GLOBALS['hdrlevel'] = Array();
$GLOBALS['emphasis'] = Array();

/* Functions required from surrounding environment:
 *   
 *  WikiLinkFunc()
 *  WikiExpressionsProcess
 *  WikiConditionsProcess
 */

function WikiStoreDyn($funcname, $text='')
{
  static $n; ++$n;
  $result = "{$n}";
  $d = Array($funcname, $text);
  if($funcname == 'SrcEmbed') $d[2] = func_get_arg(2);
  $GLOBALS['strdyn'][$n] = $d;
  return $result;
}
function WikiFetchDynd($n)
{
  $data = $GLOBALS['strdyn'][$n];
  $funcname = $data[0];
  if($funcname == 'SrcEmbed') return $funcname($data[1],$data[2]);
  return $funcname($data[1]);
}
function WikiProcessStoreDynd($result)
{
  /* Replaces the stored patterns with function call results. */
  return preg_replace("/([0-9]*)/e", 'WikiFetchDynd($1)', $result);
}

function WikiStoreConst($funcname, $text='')
{
  static $n; ++$n;
  $result = "{$n}";
  $GLOBALS['strcon'][$n] = Array($funcname, $text);
  return $result;
}
function WikiFetchConstd($n)
{
  $data = $GLOBALS['strcon'][$n];
  $funcname = $data[0];
  return $funcname($data[1]);
}
function WikiProcessStoreConstd($result)
{
  /* Replaces the stored patterns with function call results. */
  return preg_replace("/([0-9]*)/e", 'WikiFetchConstd($1)', $result);
}
function WikiIdleFunc($s)
{
  return htmlspecialchars($s);
}
function WikiLinkFuncFT($s) /* FT stands for false,true */
{
  /* cutter=false, autoembedimg=true */
  /* used for inline URLs and JoinCapitalizedWords */
  return WikiLinkFunc($s,false,true);
}

function WikiToggleEmphasis($tag)
{
  $GLOBALS['emphasis'][$tag]['aim'] ^= 1;
  return WikiRestoreEmphasis();
}
function WikiEnterEmphasis($tag)
{
  ++$GLOBALS['emphasis'][$tag]['aim'];
  return WikiRestoreEmphasis();
}
function WikiLeaveEmphasis($tag)
{
  --$GLOBALS['emphasis'][$tag]['aim'];
  return WikiRestoreEmphasis();
}
function WikiDebugEmphasis()
{
  $emdata = &$GLOBALS['emphasis'];
  ob_start();
  print '<pre>';print_r($emdata);print'</pre>';
  $result = ob_get_contents();
  ob_end_clean();
  return $result;
}
function WikiClearEmphasis()
{
  $result = '';
  
  #$result = "clear(" . WikiDebugEmphasis() . ")";
  
  $emdata = &$GLOBALS['emphasis'];
  $stack = &$emdata['stack'];
  $stacktop = &$stack['top'];
  while($stacktop > 0)
  {
    $tag = $stack[--$stacktop];
    $result .= "</$tag>";
    --$emdata[$tag]['state'];
  }
  $stack = Array();
  return $result;
}
function WikiRestoreEmphasis()
{
  $emdata = &$GLOBALS['emphasis'];
  $stack = &$emdata['stack'];
  $stacktop = &$stack['top'];
  
  $npeel = 0;
  $peel = Array();
  
  $result = '';
  
  #$result = "restore(" . WikiDebugEmphasis() . ")";
  
  /* First, check which states need to be removed */
  foreach($emdata as $tag => &$data)
  {
    if($tag == 'stack') continue;
    
    $aim   = &$data['aim'];   if($aim < 0) $aim = 0;
    $state = &$data['state'];
    if($state > $aim)
    {
      $diff = $state-$aim;
      $npeel      += $diff;
      $peel[$tag] += $diff;
    }
  }
  unset($aim); unset($state);
  
  /* Peel the active states until there's nothing that needs to be removed */
  while($stacktop > 0 && $npeel > 0)
  {
    $tag = $stack[--$stacktop];
    $result .= "</$tag>";
    
    if($peel[$tag] > 0)
    {
      --$peel[$tag];
      --$npeel;
    }
    --$emdata[$tag]['state'];
  }
  
  /* Then, activate each state that needs to be activated... in arbitrary order. */
  foreach($emdata as $tag => &$data)
  {
    if($tag == 'stack') continue;
    $aim   = &$data['aim'];
    $state = &$data['state'];
    while($state < $aim)
    {
      $result .= "<$tag>";
      $stack[(int)$stacktop++] = $tag;
      ++$state;
    }
    $state = $aim; /* consider everything has been done now */
  }
  return $result;
}
function WikiEmphasisTransition($html)
{
  return WikiStoreConst('WikiClearEmphasis') . $html . WikiStoreConst('WikiRestoreEmphasis');
}

function WikiBriefUnHTML($s)
{
  return str_replace
  (
    Array('&amp;', '&lt;', '&quot;'),
    Array('&', '<', '"'),
    $s
  );
}

function WikiCreateReference($refnum)
{
  if($GLOBALS['WIKI_PARSER'] == WIKI_PARSING_TEXT)
  {
    return '<sup>'.$refnum.'</sup>';
  }
  return '<a name="r'.$refnum.'"></a><sup>[<a href="#'.$refnum.'">'.$refnum.'</a>]</sup>';
}
function WikiCreateFootNote($refnum)
{
  if($GLOBALS['WIKI_PARSER'] == WIKI_PARSING_TEXT)
  {
    return '['.$refnum.']';
  }
  return '[<a name="'.$refnum.'"></a><a href="#r'.$refnum.'">'.$refnum.'</a>]';
}

function WikiCreateHeader($level, $label)
{
  $label = preg_replace('@^ +@', '', $label);
  
  $tmp = $GLOBALS['WIKI_PARSER'];
  $GLOBALS['WIKI_PARSER'] = WIKI_PARSING_TEXT;
   $label = str_replace(Array('&lt;', '&quot;', '&amp;', '&#91;', '&#93;'),
                        Array('<',    '"', '&', '[[', ']]'),
                        $label);
   $label_fixed = WikiTextToHTML($label); // parse out links etc
  $GLOBALS['WIKI_PARSER'] = $tmp;

  // $label is HTML which will still be wikiprocessed
  $tag = 'h'.$level;

  $anchor = '';
  WikiTOCRegisterHeader($anchor, $level, $label, $label_fixed);

  $result = '';
  
  $hdrlevel = &$GLOBALS['hdrlevel'];
  foreach(array_reverse($hdrlevel) as $saved_level)
  {
    if($saved_level < $level) break;
    array_pop($hdrlevel);
    $result .= '</div>';
  }
  
  array_push($hdrlevel, $level);
  
  $result .= '<'.$tag.'>'.
    '<a name="'.$anchor.'"></a>'. $label.
         '</'.$tag.'>';
  
  $result .= '<div class="deeper">';
  
  return WikiEmphasisTransition($result);
}
function WikiEndHeaders()
{
  $hdrlevel = &$GLOBALS['hdrlevel'];

  $result = '';
  foreach($hdrlevel as $level) $result .= '</div>';

  $hdrlevel = Array();
  return WikiEmphasisTransition($result);
}
function WikiCreateRuler()
{
  // Removing the EndHeaders() for now; it breaks %%TAB%%
  return '<hr/>';
  //return WikiEndHeaders() . '<hr/>';
}

function WikiConvertFlags($prev, $now)
{
  $min = min(strlen($now), strlen($prev));
  $max = max(strlen($prev), strlen($now));
  $prev = str_pad($prev, $max);
  $now  = str_pad($now,  $max);
  
  $result = '';
  for($common_length=0; $common_length<$max; ++$common_length)
    if($prev[$common_length] != $now[$common_length])
      break;
  
  /* $common_length = number of matching characters of format from beginning */

  #$result .= '<!-- prev(' . $prev .') now('. $now .') -->';
  
  /* Undo old markers in reverse order */
  for($c=$max; --$c>=$common_length; )
    switch($prev[$c])
    {
      case '>': $result .= '</blockquote>'; break;
      case 'T': $result .= '</table>'; break;
      case 'P': $result .= '</pre>'; break;
      case 'D': $result .= '</li></dl>'; break;
      case '*': $result .= '</li></ul>'; break;
      case '#': $result .= '</li></ol>'; break;
    }

  #if(ereg('[*#]', $prev) && substr($prev,0,$min)==substr($now,0,$min))
  #  $result .= '</li>';

  $did_li = false;
  
  /* Apply new markers */
  for($c=$common_length; $c < $max; ++$c)
    switch($now[$c])
    {
      case '>': $result .= '<blockquote>'; break;
      case 'T': $result .= '<table class="wikitable" border="1">'; break;
      case 'P': $result .= '<pre>'; break;
      case 'D': $result .= '<dl><li>'; $did_li = false; break;
      case '*': $result .= '<ul><li>'; $did_li = true; break;
      case '#': $result .= '<ol><li>'; $did_li = true; break;
    }

  if(!$did_li && preg_match('/[*#] *$/', $now))
  {
    $result .= '</li><li>';
  }

  return $result;
}

$rep1w=false;
$rep2w=false;
function WikiRealProcessMarkupLine($text)
{
  global $rep1w,$rep2w;
  if($rep1w===false)
  {
    $rep1w = Array();
    $rep2w = Array();
    foreach($GLOBALS['WIKIREPLACES'] as $a=>$b) { $rep1w[] = $a; $rep2w[] = $b; }
  }
  return preg_replace($rep1w, $rep2w, $text);
}
function WikiRealProcessMarkup($text)
{
  $result = '';
  $prev_flags = '';
  
  $maybe_p = false;
  $was_hdr = true;
  
  $text = str_replace("\r", '', $text);
  foreach(explode("\n", $text) as $line)
  {
    $line = preg_replace('/([^ ]) +$/', '\1', $line);

    if($line == '')
    {
      if(!$was_hdr) $maybe_p = true;
      continue;
    }
    if(preg_match('/^![a-z0-9 !]/', $line))
    {
      $maybe_p = false;
      $was_hdr = true;
    }
    else
    {
      if(!$was_hdr && $maybe_p && $flags != 'P')
      {
        $flags = '';
        $result .= WikiConvertFlags($prev_flags, $flags);
        $prev_flags = $flags;
        
        $result .= WikiEmphasisTransition('<p/>');
      }
      $was_hdr = false;
      $maybe_p = false;
    }
    $flags = '';
    $list = false;
    if(preg_match('/^ /', $line)) /* Preformatted text */
    {
      # $line = substr($line, 1);
      $flags = 'P';
    }
    elseif(preg_match('@^\|\|?[^|].*\|@', $line)) /* Tables */
    {
      /* ok is:
          [[
       or ]]
       or [ x ]
        where x is:
           [[
        or ]]
        or anything but [ ]
       or anything but [ ]
      */
      $linkok = '\[\[|]]|[^|[\]]';
      $okchar = $linkok.'|\[(?:'.$linkok.'|\|)+\]';
      $cells = Array();
      preg_match_all(
        '@\|{1,2}(?:'.$okchar.')+@',
        $line,
        $cells);
      $flags = 'T';
      $was_hdr = false;
    }
    elseif(preg_match('/^;.*:/', $line))
    {
      $flags = 'D';
      preg_match(
        '@^;([^:]*):(.*)@',
        $line,
        $deflist);
      $was_hdr = false;
    }
    elseif(preg_match('/^[>]+/', $line))
    {
      $backup = $line;
      $line = preg_replace('/^[>]*/', '', $line);
      $flags = utf8_decode(substr($backup, 0, -strlen($line)));
    }
    else /* Bulleted and numbered lists */
    {
      $backup = $line;
      $line = preg_replace('/^[#*]*/', '', $line);
      if($line != $backup)
      {
        $list = true;
        $flags = substr($backup, 0, -strlen($line));
      }
    }
    
    $flag_html = WikiConvertFlags($prev_flags, $flags);
    if(!empty($flag_html))
    {
      /* FIXME: results in error when <table> transitions over emphasis */
      $result .= WikiEmphasisTransition($flag_html);
    }
    if($list) { $line = preg_replace('/^ */', '', $line); }
    $prefix = ($flags == '' ? '' : ' ');
    
    if($flags == 'T')
    {
      $result .= '<tr>';
      foreach($cells[0] as $cell)
      {
        $type = 'td';
        $n=1;
        if($cell[$n] == '|') { $type = 'th'; ++$n; }

        $cell = WikiStoreConst('WikiRestoreEmphasis')
               .WikiRealProcessMarkupLine($prefix.substr($cell,$n))
               .WikiStoreConst('WikiClearEmphasis');

        $attrs = '';
        // &minus;
        if(preg_match('@^ *[-+−]*[0-9]+ *$@', $cell)) $attrs = ' align="right"';
        $result .= "<{$type}$attrs>$cell</$type>";
      }
      $result .= '</tr>';
    }
    elseif($flags == 'D')
    {
      $result .= WikiEmphasisTransition('<dt>');
      $result .= WikiRealProcessMarkupLine($prefix.$deflist[1]);
      $result .= WikiEmphasisTransition('</dt><dd>');
      $result .= WikiRealProcessMarkupLine($prefix.$deflist[2]);
      $result .= WikiEmphasisTransition('</dd>');
    }
    else
    {
      #if(ereg('^ *$', $line)) // if it's all just spaces
      #  $result .= '<br/>&nbsp;<br/>';
      #else
        $result .= WikiRealProcessMarkupLine($prefix.$line);
    }
    $result .= "\n";
    $prev_flags = $flags;
  }
  $flags = '';
  $result .= WikiConvertFlags($prev_flags, $flags);
  
  $result = WikiProcessTOC($result);
  
  $result .= WikiEndHeaders();
  $result = WikiProcessStoreConstd($result);

  $result .= WikiClearEmphasis();

  return $result;
}

function WikiTextToHTML($text)
{
  $stor1 = $GLOBALS['strdyn']; $GLOBALS['strdyn'] = Array();
  $stor0 = $GLOBALS['strcon']; $GLOBALS['strcon'] = Array();
  $toch = $GLOBALS['toc_headers']; $GLOBALS['toc_headers'] = Array();
  $hdrlevel = $GLOBALS['hdrlevel']; $GLOBALS['hdrlevel'] = Array();
  $emphasis = $GLOBALS['emphasis']; $GLOBALS['emphasis'] = Array();

  if(function_exists('WikiSrcEmbedProcess'))    $text = WikiSrcEmbedProcess($text);
  if(function_exists('WikiExpressionsProcess')) $text = WikiExpressionsProcess($text);
  if(function_exists('WikiConditionsProcess'))  $text = WikiConditionsProcess($text);

  if($GLOBALS['WIKI_PARSER'] == WIKI_PARSING_SNAPSHOT)
  {
    $result = WikiRealProcessMarkup($text);
  }
  else
  {
    $c = new BinKeyCache($text,  0, 3600);
    $data = $c->LoadOrLog();
    if(!$c->Found())
    {
      $result = WikiRealProcessMarkup($text);
      $data = Array
      (
        'result' => $result,
        'strdyn' => $GLOBALS['strdyn']
      );
      $c->Save($data);
    }
    else
    {
      $result = $data['result'];
      $GLOBALS['strdyn'] = $data['strdyn'];
    }
    $c->_Cache();
    unset($c);
  }

  $result = WikiProcessStoreDynd($result);

  if(function_exists('WikiTabsProcess')) $result = WikiTabsProcess($result);

  $GLOBALS['strdyn'] = $stor1;
  $GLOBALS['strcon'] = $stor0;
  $GLOBALS['emphasis'] = $emphasis;
  $GLOBALS['toc_headers'] = $toch;
  $GLOBALS['hdrlevel'] = $hdrlevel;

  return $result;
}

function HTMLtoWikiText($text)
{
  $text = str_replace('[', '[[', $text);
  $text = str_replace(']', ']]', $text);
  $text = preg_replace("/[\n \t]+/", ' ', $text);
  $text = preg_replace('/^ */', '', $text);
  $text = preg_replace('/ +$/', '', $text);
  
  $text = preg_replace("@ *<hr/?> *@", "\n----\n", $text);
  $text = preg_replace('@<!--.*-->@U', '', $text);
  $text = preg_replace('@<a +href="([^"]*)" *> *(.*) *< */a *>@U', '[\1|\2]', $text);
  $text = preg_replace("@ *<br */?> *@", "%%%\n", $text);
  $text = preg_replace("@ *<p/?> *@", "\n\n", $text);
  $text = str_replace('<small>', '((', $text);
  $text = str_replace('</small>', '))', $text);
  $text = str_replace('<b>', '__', $text);
  $text = str_replace('</b>', '__', $text);
  $text = str_replace('<em>', "''", $text);
  $text = str_replace('</em>', "''", $text);
  $text = str_replace('&lt;', '<', $text);
  $text = str_replace('&gt;', '>', $text);
  $text = str_replace('&quot;', '"', $text);
  $text = str_replace('&amp;', '&', $text);
  $text = str_replace('&nbsp;', ' ', $text);
  #$text = str_replace('&#91;', '[', $text);
  #$text = str_replace('&#93;', ']', $text);
  return $text;
}

function TextToWikiText($text, $wrapping=false)
{
  $text = str_replace("\r\n", "\n", $text);
  $text = str_replace("\r", '', $text);

  /* word-wrapping */
  if($wrapping)
  {
    preg_match_all("@([(]*[-\w:.!?,)]+|[^\w\n]+?|\n|.)@", $text, $words);
    $text = '';
    $len = 0;
    foreach($words[0] as $word)
    {
      if(preg_match("/\n/", $word)) { $len=0; }
      $len += strlen($word);
      if($len > 60 && !ereg('^ ', $word)) { $text .= "\r"; $len = strlen($word); }
      $text .= $word;
    }
  }

  $text = 
    preg_replace('/^([*#]+.*)%%%$/', '\1',
    preg_replace('/^- *(.+)%%%/', "* \\1\n",
   # No JoinCapitalizedWords protection needed anymore.
   # preg_replace('/((?<![a-z0-9])[A-Z]+[a-z0-9]+[A-Z][A-Za-z0-9]*)/',
   #              '!\1',
    str_replace("\r", "\n",
    str_replace("%%%\n%%%\n", "\n\n",
    str_replace("\n", "%%%\n",
    str_replace('%%%', "%%''''%'",
    str_replace('%%TOC', "%''''%TOC",
    str_replace('|', '[|]',
    str_replace('[', '[[',
    str_replace(']', ']]',
    str_replace('__', "_''''_",
    $text )))))))))));

  return $text;
}
