<?php

#require_once 'inc/layout_api.php';
require_once 'inc/smallcache.php';

function SrcEmbed($text, $syntax, $use_cache=true)
{
  $jsf_filename = "srcformats/{$syntax}.jsf";
  if(!file_exists($jsf_filename))
  {
    return htmlspecialchars("<<< JSF syntax file '$syntax' not found >>>");
  }
  $m = filemtime($jsf_filename);

  if($use_cache)
  {
    // Check for a cached HTML version
    $syncache = new BinKeyCache("rendered_jsf_$syntax:".$text, $m, 3600);
    $html = $syncache->LoadOrLog();
    if($syncache->Found())
    {
      return $html;
    }
  }
  
  require_once 'inc/jsffun.php'; // <- trying to avoid parsing it unless necessary
  
  static $jsf_table;
  $jsf = &$jsf_table[$syntax];
  if(!$jsf || !$use_cache)
  {
    $jsf = new JSF_Parser($syntax);
    $loaded = false;
    if($use_cache)
    {
      // Check for a cached JSF parsing
      $jsfcache = new Cache("saved_jsf_$syntax", $m, 3600);
      $jsfdata = $jsfcache->LoadOrLog();
      if($jsfcache->Found())
      {
        $jsf->load($jsfdata);
        $loaded = true;
      }
    }
    
    if(!$loaded)
    {
      // Parse
      $jsf->ParseSyntax(file($jsf_filename));
      
      if($use_cache)
      {
        // Save to cache
        $jsfdata = $jsf->save();
        $jsfcache->Save($jsfdata);
        unset($jsfcache);
      }
    }
  }

  ob_start();
   $style = $jsf->GetBackgroundColorInCSSformat();
  
   $class = 'source_snippet';
  
   print '<pre class="'.$class.'" style="'.htmlspecialchars($style).'">';
    $applier = new JSF_Applier($jsf);
    $applier->ApplyJSF($text);
    unset($applier);
   print '</pre>';

   $html = ob_get_contents();
   // This will also save the HTML rendering to cache.
   if($use_cache)
     $syncache->Save($html);
  ob_end_clean();
  unset($syncache);
  
  return $html;
}

function WikiSrcEmbedProcess($text)
{
  return preg_replace(
    "@%%SRC_EMBED ([a-z0-9_]+)\r?\n?(.*?)[\r\n]*%%END_EMBED\r?\n?@ise",
    "WikiStoreDyn('SrcEmbed',str_replace('\\\"', '\"', '\\2'),str_replace('\\\"', '\"', '\\1'))",
    $text);
}

function WikiLoadAllJsfStyles()
{
  require_once 'inc/jsffun.php';
  $sty = Array();
  foreach(glob('srcformats/*.jsf') as $jsffile)
  {
    $syntax = ereg_replace('\.jsf$', '', ereg_replace('.*/', '', $jsffile));
    
    $jsf = new JSF_Parser($syntax);
    $jsf->ParseSyntax(file($jsffile));
    foreach($jsf->styles as $style => $names)
      foreach($names as $name)
        $sty[$style][$name] = $name;
  }
  $jsf->styles = $sty;
  return $jsf->GetStyles();
}
